/* wmgeneral.h,v 1.2 1998/08/26 15:58:58 rneswold Exp */

#ifndef WMGENERAL_H_INCLUDED
#define WMGENERAL_H_INCLUDED

/* Typedefs */

typedef struct _rckeys rckeys;

struct _rckeys {
	const char	*label;
	char		**var;
};

typedef struct {
	Pixmap			pixmap;
	Pixmap			mask;
	XpmAttributes	attributes;
} XpmIcon;

/* Global variable */

extern Display* display;

/* Function Prototypes */

void AddMouseRegion(int index, int left, int top, int right, int bottom);
int CheckMouseRegion(int x, int y);

void openXwindow(int, char **, char const* const*, char const*, int, int);
void RedrawWindow(void);
void RedrawWindowXY(int x, int y);

void copyXPMArea(int, int, int, int, int, int);
void copyXBMArea(int, int, int, int, int, int);
void setMaskXY(int, int);

#endif
